/*
 * Created on May 3, 2005
 *
 * Last Modified on Feb 24, 2007
 * Version: 0.5
 */
/**
 * @author zAlbee
 */
import java.io.*;
import java.util.*;
import java.text.*;
import com.freeware.inifiles.INIFile;

public class immerge {

	static final String NL = System.getProperty("line.separator"); // newline (CR/CRLF/LF)

	// The Byte-Order Mark (BOM), when occuring at the beginning of a file, indicates a Unicode file 
	// It is the single Unicode character U+FEFF (ZERO WIDTH NO-BREAK SPACE)
	// Generally, it is a convention of Win32 programs to place the BOM, but not POSIX systems.
	static final char UNICODEBOM = 0xFEFF; // BOM as Unicode character
	static final byte[] UTF8BOM = {(byte)0xEF, (byte)0xBB, (byte)0xBF}; // BOM encoded in UTF-8

	static final String INI_SECTION_ICQNAMES = "ICQ Names";
	static final int AIM = 0;
	static final int ICQ = 1;
	static final int IRC = 2;
	static final int MSN = 3;
	static final int YAHOO = 4;
	
	static int tz_adjust = -16;
	
	static INIFile settingsINI;
	
	/***
	 * Log abstract class, associated with a BufferedReader
	 *
	 * Subclasses need to implement:
	 *
	 *   readSession() - reads the next session into a buffer
	 *              - puts that session's date into the sessDate property.
	 *              - returns True if a session was read, False if the log was empty (no session read)
	 *
	 *   getSession() - should return all the text within that session.
	 */
	static abstract class Log implements Comparable {
		static final String sessStart  = "Session Start (";
		static final String sessStart2 = "sion Start (";
		static final String sessClose  = "Session Close (";
		static SimpleDateFormat sdfTrillian = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy", Locale.US);
		static SimpleDateFormat sdfICQ = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss", Locale.US);
		static SimpleDateFormat sdfTTime = new SimpleDateFormat("[hh:mm a]", Locale.US);
		static SimpleDateFormat sdfTTime24 = new SimpleDateFormat("[HH:mm]", Locale.US);

		String filename;
		String pathname;
		BufferedReader reader;
		Date sessDate;
		String buffer;
		boolean empty = false;
		int priority = 0; // when multiple logs have the same session date, the higher priority log's session is used.

		abstract public boolean readSession() throws Exception;
		abstract public String getSession();

		/** Log() constructor - automatic UTF-8 detection
		 */
		Log(File file) throws IOException {

			filename = file.getName();
			pathname = file.getPath();
			
			reader = new BufferedReader(new FileReader(file));
			
			reader.mark(16);
			if (reader.read() == 0xEF && reader.read() == 0xBB && reader.read() == 0xBF) { 
				// Detected UTF-8 header:
				// Use UTF-8 reader, but we still need to ignore the first 3 bytes (1 character).
				
				InputStreamReader in = new InputStreamReader(new FileInputStream(file), "utf-8");
				reader = new BufferedReader(in);
				
				reader.mark(16);
				if (reader.read() == 0xFEFF) {
				}
				else {
					// As long as Java's UTF-8 reader works, this should never happen
					// Maybe if the system is byte-swapped.
					System.out.println("Warning: UTF-8 BOM was detected, but was not treated as such. Send a bug report to zalbee" + "@gmail.com");
					reader.reset();
				}
			}
			else reader.reset();
		}

		/** Log() constructor - explicit charset
		 */
		Log(File file, String charsetName) throws IOException {
			
			filename = file.getName();
			pathname = file.getPath();
			
			InputStreamReader in = new InputStreamReader(new FileInputStream(file), charsetName);
			reader = new BufferedReader(in);
				
			reader.mark(16);
			if (in.getEncoding() == "UTF8" && reader.read() == 0xFEFF) {
			}
			else {
				reader.reset();
			}
		}
		
		/** Allows sorting instances of Log by date and priority.
		 */
		public int compareTo(Object o2) {
			Log log2 = (Log)o2;
			int cmp = sessDate.compareTo(log2.sessDate);
			if (cmp == 0) return log2.priority - priority;
			else return cmp;
		}

	} // end abstract class Log
	
	
	/***
	 * ConverterLog abstract class, subclass of Log
	 *
	 *   This is for log formats that store history in a different way than Trillian.
	 *   This class has methods for converting single messages to Trillian-style messages and sessions.
	 *   Most non-Trillian log formats (eg. ICQ, MSN) will use this class - TrLog and TrXML do NOT
	 *   inherit this class because they directly copy and do not convert the messages.
	 *
	 * Defines methods:
	 *
	 *   readSession() - Reads the next session into buffer. A session is multiple messages within a certain timeframe.
	 *   getTrMsg()     - Returns a Trillian-log formatted version of the stored message.
	 *   getTrSession() - Returns a Trillian-log formatted version of the stored session.
	 *   getXMLMsg()     - Returns a Trillian-XML formatted version of the stored message.
	 *   getXMLSession() - Returns a Trillian-XML formatted version of the stored session.
	 * 
	 * Subclasses need to implement:
	 *
	 *   readMsg() - reads the next message (as plaintext, with no entities, and no extra meta-information) into msgBuffer
	 *             - msgBuffer should not have an extra newline character at the end (unless that is part of the message)
	 *             - returns true if a message was read, otherwise false
	 *             - sets msgIsSend = true if that message is sent from the owner of this account 
	 *               (yourNick), otherwise false
	 *             - can also set properties yourNick and nickName
	 *
	 *   getMsg() - suggested to point to one of getTrMsg() or getXMLMsg()
	 *
	 *   getSession() - suggested to point to one of getTrSession() or getXMLSession()
	 *
	 *   Note: If you plan to override the readSession() method, then it is not necessary to follow the above guidelines
	 *   for readMsg(), getMsg(), and getSession().
	 *
	 */
	static abstract class ConverterLog extends Log {
		protected String msgBuffer;
		protected Date msgDate;
		protected boolean msgIsSend;

		protected String msgBuffer2;
		protected Date msgBuffer2Date = null;

		protected Date sessEndDate;
		
		// It is up to the subclass to set these:
		protected String medium;
		protected String userID;
		protected String yourUserID;
		protected String nickName;
		protected String yourNick;
		
		public void setUserID(String theirUIN) {
			this.userID = theirUIN;
		}
		
		public void setYourUserID(String yourUIN) {
			this.yourUserID = yourUIN;
		}

		public void setNick(String theirNick) {
			this.nickName = theirNick;
		}

		public void setYourNick(String yourNick) {
			this.yourNick = yourNick;
		}

		public String getUserID() {
			return userID;
		}

		public String getNick() {
			return nickName;
		}

		public String getYourNick() {
			return yourNick;
		}

		public String getYourUserID() {
			return yourUserID;
		}

		public ConverterLog(File file) throws IOException {
			super(file);
		}

		public ConverterLog(File file, String charsetName) throws IOException {
			super(file, charsetName);
		}
		
		abstract protected boolean readMsg() throws Exception;
		abstract protected String getMsg();

		/** ConverterLog.readSession()
		 *   Reads one session and stores it.
		 *
		 *   This method calls readMsg() and getMsg() multiple times, and concatenates the returned Strings into one
		 *   session String. The number of messages in a single session equals the number of messages
		 *   that are dated within 24 hours of the first message in that session.
		 */
		public boolean readSession() throws Exception {
			sessDate = msgBuffer2Date;
			String msg = null;

			if (sessDate == null) {
				boolean read = readMsg();
				if (!read) {empty=true; reader.close(); return false;}
				msg = getMsg();
				sessDate = msgDate;
			}
			else {
				msg = msgBuffer2;
				msgBuffer2 = null;
				msgBuffer2Date = null;
			}

			sessEndDate = sessDate;
			buffer = msg;

			boolean hasMsg = readMsg();

			while (hasMsg) {
				msg = getMsg();

				if (msgDate.getTime() < sessDate.getTime() + 24*60*60*1000) {
					buffer += msg;
					sessEndDate = msgDate;
				}
				else {
					msgBuffer2 = msg;
					msgBuffer2Date = msgDate;
					break;
				}

				hasMsg = readMsg();
			}

			return true;
		}

		/** ConverterLog.getTrMsg()
		 *   Returns a Trillian-log formatted version of the stored message.
		 */
		protected String getTrMsg() {
			String prefix = sdfTTime.format(msgDate) + " ";
			if (msgIsSend) prefix += yourNick + ": ";
			else prefix += nickName + ": ";
			
			String suffix = NL;

			String tmpMsg = msgBuffer.replaceAll("\n", "\n ");
			
			return prefix + tmpMsg + suffix;
		}

		/** ConverterLog.getTrSession()
		 *   Returns a Trillian-log formatted version of the stored session.
		 */
		protected String getTrSession() {
			String prefix = sessStart + (yourUserID.length()>0 ? yourUserID : medium) + ":" + nickName + "): " + sdfTrillian.format(sessDate) + NL;
			String suffix = sessClose + nickName + "): " + sdfTrillian.format(sessEndDate) + NL+NL+NL;

			return prefix + buffer + suffix;
		}

		/** ConverterLog.getXMLMsg()
		 *   Returns a Trillian-XML formatted version of the stored message.
		 */
		protected String getXMLMsg() {
			//<message type="incoming_privateMessage" time="1132707630" medium="MSN" to="zalbee2%40hotmail%2Ecom" from="catty_kitty_leung%40hotmail%2Ecom" from_display="%28%40%29%20SuPeR%20cAt%20%2D%20Jay%20Chow%20is%20amazing" text="that%20was%20a%20FAST%20response"/>
			//<message type="outgoing_privateMessage" time="1133042238" medium="ICQ" to="16225800" from="18196095" from_display="zAlbee" text="hm%2E"/>
			String prefix = "<message type=\"" + (msgIsSend ? "outgoing_privateMessage" : "incoming_privateMessage")
					+ "\" time=\"" + msgDate.getTime()/1000 + "\" medium=\"ICQ\" to=\""
					+ (msgIsSend ? userID : yourUserID) + "\" from=\"" + (msgIsSend ? yourUserID : userID)
					+ "\" from_display=\"" + (msgIsSend ? yourNick : nickName) + "\" text=\"";
			
			String suffix = "\"/>" + NL;
			
			// in the XML, <a href (more correctly, %3Ca%20href) starts a link in the history.
			// if chat uses <a href or &, must be html entitified to prevent unintended behaviour; > is ok as %3E
			String tmpMsg = msgBuffer.replaceAll("\n", "\n ").replaceAll("&","&amp;").replaceAll("<","&lt;");
			
			return prefix + numericEntity(tmpMsg) + suffix;
		}

		/** ConverterLog.getXMLSession()
		 *   Returns a Trillian-XML formatted version of the stored session.
		 */
		protected String getXMLSession() {
			//<session type="start" time="1133041383" medium="ICQ" to="16225800" from="18196095"/>
			String prefix = "<session type=\"start\" time=\"" + sessDate.getTime()/1000 +  "\" medium=\"ICQ\" to=\"" 
					+ userID +"\" from=\""+ yourUserID + "\"/>" + NL;

			//<session type="stop" time="1133060545" medium="ICQ" to="16225800" from="18196095"/>
			String suffix = "<session type=\"stop\" time=\"" + sessEndDate.getTime()/1000 +  "\" medium=\"ICQ\" to=\"" 
					+ userID +"\" from=\""+ yourUserID + "\"/>" + NL;
					
			return prefix + buffer + suffix;
		}
	} // end abstract class ConverterLog


	/***
	 * ICQHRLog class, subclass of ConverterLog
	 * Represents an ICQ log file (ICQ History Reader HTML format)
	 */
	static class ICQHRLog extends ConverterLog {
		protected String medium = "ICQ Pro";
		
		private String sendColour;
		private String recvColour;
		private static final String genStart = "<font size=3 color=";
		private static final String genStart2 = ">";
		private static final String genEnd = "</font>";
		private static final int START_LEN = genStart.length() + 8;
		private String sendStart;
		private String recvStart;
		
		/** ICQHRLog() contructor
		 */
		public ICQHRLog(File f) throws IOException {
			super(f);
			priority = 2;
		}
		
		/** ICQHRLog.init()
		 * Sets userID, yourUserID, yourNick. Tries to find nickName from log.
		 * Prepares log for reading.
		 */
		public void init() throws IOException{
			
			String s = reader.readLine();

			while (s != null) {
				if (s.matches("<td [^>]*>Nickname:</td>")) {
					s=reader.readLine();
					nickName = s.substring( s.indexOf('>') +1, s.lastIndexOf("</td>") );
					break;
				}
				else if (s.trim().toLowerCase().equals("no info for this user")) {
					System.out.println("Info: No nickname in " + filename);
					nickName = userID;
					s = reader.readLine();
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Reached end-of-file while looking for info in " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Sent</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					sendColour = s.substring( colIdx, colIdx + 7);
					sendStart = genStart + sendColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read sent colour from " + filename);
				empty = true;
				return;
			}

			while (s != null) {
				if (s.matches("<tr><td [^>]*>Recieved</td><td.*bgcolor=.*")) {
					int colIdx = s.indexOf("bgcolor=") + 8;
					recvColour = s.substring( colIdx, colIdx + 7);
					recvStart = genStart + recvColour + genStart2;
					break;
				}
				s=reader.readLine();
			}

			if (s == null) {
				System.out.println("Error: Could not read receive colour from " + filename);
				empty = true;
				return;
			}

		}

		/** ICQHRLog.readMsg()
		 *    returns true if a message was read, otherwise false
		 *    use getMsg() to see the results of this operation
		 */
		protected boolean readMsg() throws Exception {
			boolean msgOpen = false;
			msgBuffer = "";

			String s=reader.readLine();
			msgDate = null;

			// grab date

			while (s != null && msgDate==null) {
				if ( s.startsWith(genStart) ) {
					int dateStartIdx = s.indexOf('(', START_LEN) + 1;
					int dateEndIdx = s.lastIndexOf(')');
					msgDate = sdfICQ.parse(s.substring(dateStartIdx, dateEndIdx));
					//dpoint[0].setTime( dpoint[0].getTime() - 16*60*60*1000 ); // has sync problems when dates are near daylight savings switch
					msgDate.setHours( msgDate.getHours() + tz_adjust); // -16 hours seems to make the time correct? no idea how ICQHR calculates timezone
				}
				s = reader.readLine();
			}
			
			// grab colour (to determine sender) and message

			while (s != null && !msgOpen) {
				if ( s.startsWith(genStart)) {
					msgOpen = true;
					msgIsSend = s.startsWith(sendStart);

					if (s.trim().endsWith(genEnd)) {
						msgBuffer += unHTMLEntity( s.substring( START_LEN, s.lastIndexOf(genEnd) ) );
						msgOpen = false;
						return true;
					}//break;}
					else if (s.trim().endsWith("<BR>")) msgBuffer += unHTMLEntity( s.substring( START_LEN, s.lastIndexOf("<BR>") ) ) + NL;
					else msgBuffer += unHTMLEntity( s.substring( START_LEN ) ) + NL;
				}

				s=reader.readLine();
			}

			// if message consists of more than one line

			while (s != null && msgOpen) {
				if (s.trim().endsWith(genEnd)) {
					msgBuffer += unHTMLEntity( s.substring( 0, s.lastIndexOf(genEnd)) );
					msgOpen = false;
					return true;
				} //break;}
				else if (s.trim().endsWith("<BR>")) msgBuffer += unHTMLEntity( s.substring( 0, s.lastIndexOf("<BR>") ) ) + NL;
				else msgBuffer += unHTMLEntity(s) + NL;

				s=reader.readLine();
			}
			
			// if we're here, message didn't terminate, OR there was no message in the first palce
			
			if (msgOpen) return true;
			else return false;

			//return msg;
		}
		
		/** ICQHRLog.getMsg()
		 */
		protected String getMsg() {
			return getTrMsg();
		}

		/** ICQHRLog.getSession()
		 */
		public String getSession() {
			return getTrSession();
		}

	} //end class ICQHRLog
	
	
	/***
	 * ICQLiteLog class, subclass of ConverterLog
	 * Represents an ICQ Lite log file
	 */
	static class ICQLiteLog extends ConverterLog {
		protected String medium = "ICQ Lite";
		
		/** ICQLiteLog() contructor
		 */
		public ICQLiteLog(File f) throws IOException {
			super(f, "utf-8");
			priority = 2;
		}

		/** ICQLiteLog.readMsg()
		 *    returns true if a message was read, otherwise false
		 *    use getMsg() to see the results of this operation
		 */
		protected boolean readMsg() throws Exception {
			boolean eventOpen = false;
			boolean textOpen = false;
			boolean hasMsg = false;
			
			msgBuffer = "";

			String s=reader.readLine();
			msgDate = null;
			
			// parse XML. with XML, tags can occur in any order.
			
			while (s != null && !eventOpen) {
				if (s.trim().equals("<event>")) {
					eventOpen = true;
				}
				s = reader.readLine();
			}
			
			while (s != null && eventOpen) {
				String str = s.trim();
				if (textOpen) {
					if (s.startsWith("</text>")) { // this is the expected case
						textOpen = false;
						continue;
					}
					else if (s.endsWith("</text>")) { // this does not happen AFAIK
						textOpen = false;
						msgBuffer += NL + unHTMLEntity2(s.substring(0, s.indexOf("</text>")));
					}
					else {
						msgBuffer += NL + unHTMLEntity2(s);
					}
				}
				else if (s.indexOf("<text>") != -1) {
					int pos = s.indexOf("<text>") + 6;
					msgBuffer = unHTMLEntity2(s.substring(pos));
					textOpen = true;
					hasMsg = true;
				}
				else if (str.equals("</event>")) {
					eventOpen = false;
					// must break here, to prevent extra read
					if (hasMsg) break;
				}
				else if (str.startsWith("<incoming>Yes")) {
					msgIsSend = false;
				}
				else if (str.startsWith("<incoming>No")) {
					msgIsSend = true;
				}
				else if (str.startsWith("<time>")) {
					String secs = str.substring(6, str.indexOf("</"));
					msgDate = new Date(Long.parseLong(secs) * 1000);
				}
				s = reader.readLine();
			}
			
			return hasMsg;
		}
		
		/** ICQLiteLog.getMsg()
		 */
		protected String getMsg() {
			return getTrMsg();
		}

		/** ICQLiteLog.getSession()
		 */
		public String getSession() {
			return getTrSession();
		}

	} //end class ICQLiteLog


	/***
	 * ICQHRtoXML class, subclass of ICQHRLog
	 * Represents the XML version of the ICQ History Reader file
	 */
	static class ICQHRtoXML extends ICQHRLog {
		/** ICQHRtoXML() contructor
		 */
		public ICQHRtoXML(File f) throws IOException {
			super(f);
			priority = 2;
		}
		
		/** ICQHRLog.getMsg()
		 */
		protected String getMsg() {
			return getXMLMsg();
		}

		/** ICQHRLog.getSession()
		 */
		public String getSession() {
			return getXMLSession();
		}
	}


	/***
	 * ICQLitetoXML class, subclass of ICQLiteLog
	 * Represents the XML version of the ICQ Lite log file
	 */
	static class ICQLitetoXML extends ICQLiteLog {
		/** ICQLitetoXML() contructor
		 */
		public ICQLitetoXML(File f) throws IOException {
			super(f);
			priority = 2;
		}
		
		/** ICQLiteLog.getMsg()
		 */
		protected String getMsg() {
			return getXMLMsg();
		}

		/** ICQLiteLog.getSession()
		 */
		public String getSession() {
			return getXMLSession();
		}
	}


	/***
	 * TrLog class, subclass of Log
	 * Represents a Trillian logfile.
	 */
	static class TrLog extends Log {
		private boolean sessOpen = false;
		private boolean startOfFile = true;

		/** TrLog() contructor
		 */
		public TrLog(File f) throws IOException {
			super(f);
		}

		/** TrLog.readSession()
		 */
		public boolean readSession() throws Exception {
			sessOpen = false;
			String s = reader.readLine();

			int nameEndIdx;

			while (s != null && !sessOpen) {
				if (    ( s.startsWith( sessStart ) || (startOfFile && s.startsWith(sessStart2)) )
					 && ( nameEndIdx=s.indexOf("):", s.length()-28) ) > -1
					) {
					int dateStartIdx = nameEndIdx + 3;

					sessDate = sdfTrillian.parse(s.substring(dateStartIdx));

					if (s.startsWith("sion")) {
						s = "Ses" + s;
					}
					sessOpen = true;
					buffer = s + NL;
				}
				else if (s.trim().length()>0) {
					// Text exists outside a session. Write it anyway to keep fidelity of logs.
					// Give it a "dummy" session, with the same date as the last one, so that it's written right away.
					// Since there will be no session close marker, Read until the next session starts, then rewind the reader.
					
					/*
					Calculation of max rewind for   BufferedReader.mark(int readAheadLimit)
					
					"Session Start (email@hotmail.com:MY LONG MSN DISPLAY NAME): Sun Jul 25 00:54:39 2004"
					"Session Start (".length() == 15;
					"): Sun Jul 25 00:54:39 2004".length() == 27
					Email address max == 64 + '@' + 255
						  		  	  == 320
					MSN name max == 129;
					----
					Total == 15 + 320 + ":" + 129 + 27
						  == 492
					*/
					
					buffer = s + NL;
					reader.mark(512);
					s = reader.readLine();
					
					while (s != null) {
						if ( s.startsWith( sessStart ) && s.indexOf("):", s.length()-28) > -1 ) {
							reader.reset();
							break;
						}
						buffer += s + NL;
						reader.mark(512);
						s = reader.readLine();
					}
					
					startOfFile = false;
					return true;
					
					//System.out.println("Warning: Skipped line = " + s);
				}
				
				startOfFile = false;
				s = reader.readLine();
			}

			while (s != null && sessOpen) {
				buffer += s + NL;

				if ( s.startsWith( sessClose ) && ( nameEndIdx=s.indexOf("):", s.length()-28) ) > -1 ) {
			  		sessOpen = false;
			  		buffer += NL + NL;
			  		break;
				}
				s = reader.readLine();
			}

			if (s==null) {
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			return true;

		}

		/** TrLog.getSession()
		 */
		public String getSession() {
			return buffer;
		}
	} //end class TrLog


	/***
	 * TrXML class, subclass of Log
	 * Represents a Trillian XML log file (in the "query" folder)
	 */
	static class TrXML extends Log {
		private boolean sessOpen = false;
		
		/** TrXML() contructor
		 */
		public TrXML(File f) throws IOException {
			super(f);
			priority = 1;
		}

		public boolean readSession() throws Exception {
			sessOpen = false;
			String s = reader.readLine();

			int nameEndIdx;

			while (s != null && !sessOpen) {
				if ( s.toLowerCase().startsWith( "<session type=\"start\"" ) ) {
					
					int dateStartIdx = s.indexOf("time=\"") + 6;
					int dateEndIdx = s.indexOf("\"", dateStartIdx);
					
					sessDate = new Date( 1000 * Long.parseLong( s.substring(dateStartIdx, dateEndIdx) ) );
					
					sessOpen = true;
					buffer = s + NL;
				}
				else if (s.trim().length()>0) {
					//System.out.println("Warning: skipping line: " + s);
					buffer = s + NL;
					reader.mark(1024);
					s = reader.readLine();
					
					while (s != null) {
						if ( s.toLowerCase().startsWith( "<session type=\"start\"" ) ) {
							reader.reset();
							break;
						}
						buffer += s + NL;
						reader.mark(1024);
						s = reader.readLine();
					}
					return true;
				}
				
				s = reader.readLine();
			}

			while (s != null && sessOpen) {
				buffer += s + NL;

				if ( s.toLowerCase().startsWith( "<session type=\"stop\"" ) ) {
			  		sessOpen = false;
			  		break;
				}
				s = reader.readLine();
			}

			if (s==null) {
				if (sessOpen) ; // not terminated correctly. write it anyway.
				else {
					empty = true;
					reader.close();
					return false;
				}
			}
			return true;
			
		}
		
		public String getSession() {
			return buffer;
		}
		
	} //end class TrXML


	/**
	 * unHTMLEntity ( String )
	 *
	 * Returns a new string with various HTML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 * 
	 * This one is specific for ICQ History Reader.
	 */
	static String unHTMLEntity(String s) {
		return s.replaceAll("&nbsp;"," ").replaceAll("&lt;","<").replaceAll("&gt;",">").replaceAll("&amp;","&");
	}
	
	/**
	 * unHTMLEntity2 ( String )
	 *
	 * Returns a new string with various XML escape sequences (e.g. &amp;) removed
	 * and replaced with the normal characters.
	 * 
	 * This one is specific for ICQ Lite.
	 */
	static String unHTMLEntity2(String s) {
		return s.replaceAll("&lt;","<").replaceAll("&gt;",">").replaceAll("&quot;","\"").replaceAll("&amp;","&");
	}
	
	/**
	 * numericEntity ( String )
	 *
	 * Returns a new string with all characters not in the set [0-9A-Za-z_] replaced with %hh
	 * where hh is the character's byte value in hex.
	 *
	 * Generally hh is two digits, as most chars are 1-byte. However Java String natively supports 
	 * 2-byte Unicode chars, in which case you may get %hhh or %hhhh.
	 *
	 * This function is equivalent to rawurlencode() in PHP for all 1-byte characters.
	 */
	 
	static String numericEntity(String s) {
		// make sure we escape unicode characters correctly - 
		// the way it is done, is to encode in UTF-8, then escape each byte in that encoding with %hh notation
		ByteArrayOutputStream byteStream;
		OutputStreamWriter sw;
		byte[] bytes;
		byte b;
		char c;
		String hh;
		
		byteStream = new ByteArrayOutputStream();
		try {
			sw = new OutputStreamWriter(byteStream, "utf8");
			sw.write(s);
			sw.close();
		} catch (UnsupportedEncodingException e) {
			System.out.println("Error: utf8 encoding not supported.");
		} catch (IOException e) {
			e.printStackTrace();
		}
		bytes = byteStream.toByteArray();

		// String out = "";
		StringBuffer out = new StringBuffer();

// 		for (int i=0; i<s.length(); i++) {
// 			c = s.charAt(i);
		for (int i=0; i<bytes.length; i++) {
			b = bytes[i];
			if ( ('0'<=b && b<='9') || ('a'<=b && b<='z') || ('A'<=b && b<='Z') || b=='_' ) {
				out.append((char)b);
			}
			else {
				hh = Integer.toHexString(0xff & b).toUpperCase();
				if (hh.length() < 2) hh = "0"+hh;
				else if (hh.length() > 2) System.out.println("Warning: Character did not become proper entity " + (char)b + "=" + hh);
				out.append("%"+hh);
			}
		}
		return out.toString();
	}

	/**
	 * readIntFrom ( BufferedReader )
	 *
	 * Reads an integer from the stream, returns it.
	 * Will reprompt and retry the read until an integer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static int readIntFrom(BufferedReader reader) throws IOException{
		String utxt = reader.readLine();
		try {
			int i = Integer.parseInt(utxt.trim());
			return i;
		} catch (Exception e) {
			System.out.print("You must enter a number: ");
			return readIntFrom(reader);
		}
	}
	
	/**
	 * readYesNoFrom ( BufferedReader )
	 *
	 * Reads the stream for Yes or No, returning True or False, respectively.
	 * Will reprompt and retry the read until an answer is found.
	 * Throws IOException if an I/O error occurs.
	 */
	static boolean readYesNoFrom(BufferedReader reader) throws Exception{
		boolean yesno = false;
		String utxt = reader.readLine().trim() + " ";
		char first = utxt.charAt(0);
		if (first == 'Y' || first == 'y') return true;
		else if (first == 'N' || first == 'n') return false;
		else {
			System.out.print("Yes or No? ");
			return readYesNoFrom(reader);
		}
	}

	/***
	 *  main ( String[] )
	 *
	 *  Asks the user for some information.
	 *  Then executes the merging of Trillian logs.
	 ***/
	public static void main(String[] args) throws Exception {
		
		int numTrLog, numHtml, numICQLite = 0;
		int yourUIN = 0;
		String yourNick = null;

		String[] mediums = {"AIM", "ICQ", "IRC", "MSN", "YAHOO"};
		String outPath = "output";

		File outDir = new File(outPath);
		if (!outDir.isDirectory() && !outDir.mkdir()) {
			System.out.println("Couldn't create '" + outPath + "' directory");
			System.out.println("Failed.");
			return;
		}
		
		settingsINI = new INIFile("immerge.ini");

		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		System.out.print("How many folders (logs1 .. logsN) of Trillian-style <username>.log: ");
		numTrLog = readIntFrom(in);
		System.out.print("How many folders (icqh1 .. icqhN) of IcqHR output <UIN>.html (UIN per file): ");
		numHtml = readIntFrom(in);
		System.out.print("How many folders (icql1 .. icqlN) of ICQ Lite history: ");
		numICQLite = readIntFrom(in);

		if (numHtml > 0 || numICQLite > 0) {
			System.out.print("Your ICQ UIN/number (required): ");
			yourUIN = readIntFrom(in);
			System.out.print("Your ICQ Nickname (as you want it to appear): ");
			yourNick = in.readLine();
			System.out.print("Display time in 24-hr format? (yes = 24-hr, no = AM/PM): ");
			boolean use24h = readYesNoFrom(in);
			if (use24h) {
				Log.sdfTTime = Log.sdfTTime24;
			}
		}

		if (numHtml > 0) {
			System.out.print("Correct ICQHR timezone by how many hours? (-16 = default, 0 = no change): ");
			tz_adjust = readIntFrom(in);
		}
		
		System.out.print("Should the logs be output to a \"Query\" or \"Channel\" folder like Trillian 3? (Yes/No): ");
		boolean useQuery = readYesNoFrom(in);
		
		System.out.println();

		for (int j=0; j<mediums.length; j++) {
			System.out.println("== " + mediums[j] + " ==");

			// link the username to a Vector of the logs
			Hashtable userLogs = new Hashtable();

			// link the username to a Vector of the XML files
			Hashtable userXMLs = new Hashtable();

			// Find files - Trillian logs
			for (int i=0; i<numTrLog; i++) {
				String logsPath = "logs" + (i+1) + File.separator + mediums[j]; // in folder
				String queryPath = logsPath + File.separator + "Query";         // in Query folder
				String channelPath = logsPath + File.separator + "Channel";     // in Channel folder
				
				File dir = new File(logsPath);
				File dirQ = new File(queryPath);
				File dirC = new File(channelPath);
				if (!dir.isDirectory()) continue;

				File[][] files = new File[3][];
				files[0] = dir.listFiles();
				files[1] = dirQ.isDirectory() ? dirQ.listFiles() : new File[0];
				files[2] = dirC.isDirectory() ? dirC.listFiles() : new File[0];

// 				if (dirQ.isDirectory())
// 					files[1] = dirQ.listFiles();
// 				else
// 					files[1] = new File[0];
					
				for (int qd = 0; qd <= 2; qd++) {
					for (int k=0; k<files[qd].length; k++) {
						File f = files[qd][k];
						String filename = f.getName();

						// Trillian .log
						if (f.isFile() && filename.endsWith(".log")) {
							TrLog lg = new TrLog( f );
							String user = filename.substring(0, filename.length()-4);
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}
						}

						// Trillian .xml
						if (qd!=0 && f.isFile() && filename.endsWith(".xml")) {
							
							if (filename.endsWith("-assets.xml")) {
								// assets, don't process - Trillian finds automatically
							}
							else {
								TrXML xml = new TrXML( f );
								String user = filename.substring(0, filename.length()-4);
								if (userXMLs.containsKey(user))
									((Vector)userXMLs.get(user)).add(xml);
								else {
									Vector XMLs = new Vector();
									XMLs.add(xml);
									userXMLs.put(user, XMLs);
								}
							}
							
						}
						
					} // k (files)
				} // qd
				
			} // i (numTrLog)
		
			// Find files - ICQ History Reader 1.8 .html output
			if (j==ICQ) { // ICQ

				for (int i=0; i<numHtml; i++) {
					File dir = new File("icqh" + (i+1));
					if (!dir.isDirectory()) continue;

					File[] files;
					files = dir.listFiles();
					for (int k=0; k<files.length; k++) {
						String filename = files[k].getName();

						if (files[k].isFile() && filename.endsWith(".html")) {
							String user = filename.substring(0, filename.length()-5);
							ICQHRLog lg = new ICQHRLog( files[k] );
							lg.setUserID(user);
							lg.setYourUserID(""+yourUIN);
							lg.setYourNick(yourNick);
							lg.init();
							
							// See if the init() function found a nickname, if so, save it.
							String nick = lg.getNick();
							if (nick != user) {
								settingsINI.setStringProperty(INI_SECTION_ICQNAMES, user, nick, null);
							}
							
							if (userLogs.containsKey(user))
								((Vector)userLogs.get(user)).add(lg);
							else {
								Vector logs = new Vector();
								logs.add(lg);
								userLogs.put(user, logs);
							}

							ICQHRtoXML xml = new ICQHRtoXML( files[k] );
							xml.setUserID(user);
							xml.setYourUserID(""+yourUIN);
							xml.setYourNick(yourNick);
							xml.init();
							if (userXMLs.containsKey(user))
								((Vector)userXMLs.get(user)).add(xml);
							else {
								Vector XMLs = new Vector();
								XMLs.add(xml);
								userXMLs.put(user, XMLs);
							}

						}
					} // k (files)

				} // i (numHtml)
			} // if j==1 (icq)
			
			// Find files - ICQ Lite logs
			if (j==ICQ) { // ICQ

				for (int i=0; i<numICQLite; i++) {
					File dir = new File("icql" + (i+1));
					if (!dir.isDirectory()) continue;

					// inside should be many directories
					File[] subDirs;	
					subDirs = dir.listFiles();
					for (int k=0; k<subDirs.length; k++) {
						File subDir = subDirs[k];
						if (!subDir.isDirectory()) continue;
						String user = subDir.getName(); // their ICQ number
						
						// get nickname from INI file
						String nickName = settingsINI.getStringProperty(INI_SECTION_ICQNAMES, user);
						if (user.equals("UnreadMsg")) {
							nickName = user;
						}
						else if (nickName == null) {
							System.out.print("Enter the nickname for user " + user + ": ");
							nickName = in.readLine();
							if (nickName.length() > 0) {
								settingsINI.setStringProperty(INI_SECTION_ICQNAMES, user, nickName, null);
							}
							else {
								nickName = user;
							}
						}
						
						File[] files = subDir.listFiles();
						
						for (int m=0; m<files.length; m++) {
							File file = files[m];
							String filename = file.getName();

							if (file.isFile() && filename.endsWith(".xml")) {
								ICQLiteLog lg = new ICQLiteLog( file );
								lg.setUserID(user);
								lg.setNick(nickName);
								lg.setYourUserID(""+yourUIN);
								lg.setYourNick(yourNick);

								if (userLogs.containsKey(user))
									((Vector)userLogs.get(user)).add(lg);
								else {
									Vector logs = new Vector();
									logs.add(lg);
									userLogs.put(user, logs);
								}
	
								ICQLitetoXML xml = new ICQLitetoXML( file );
								xml.setUserID(user);
								xml.setNick(nickName);
								xml.setYourUserID(""+yourUIN);
								xml.setYourNick(yourNick);
								
								if (userXMLs.containsKey(user))
									((Vector)userXMLs.get(user)).add(xml);
								else {
									Vector XMLs = new Vector();
									XMLs.add(xml);
									userXMLs.put(user, XMLs);
								}
	
							}
						} // m (files)
					} // k (subDirs)

				} // i (numICQLite)
			} // if j==ICQ


			String outPath2 = outPath + File.separator + mediums[j];
			String outPathQuery = outPath2 + File.separator + "Query";
			String outPathChannel = outPath2 + File.separator + "Channel";


			/** ---- merge Logs ---- **/
			
			// if logs exist, make output dir
			if (userLogs.size() > 0) {
				System.out.println("Starting merge, output format: Trillian Log\n");

				//outPath2[j] = outPath + File.separator + mediums[j] + (useQuery? File.separator+"Query" : "");
				
				String outPath3 = outPathQuery;
				String outPath3Ch = outPathChannel;
				
				if (!useQuery) {
					File outDir2 = new File(outPath2);
					if (!outDir2.isDirectory() && !outDir2.mkdirs()) {
						System.out.println("Error: Couldn't create directory '" + outPath2 + "'");
						System.out.println("Skipping " + mediums[j]);
						continue;
					}
					outPath3 = outPath3Ch = outPath2;
				}
				else {
					File outDir3 = new File(outPathQuery);
					if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
						System.out.println("Error: Couldn't create directory '" + outPathQuery + "'");
						System.out.println("Skipping " + mediums[j]);
						continue;
					}
					outDir3 = new File(outPathChannel);
					if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
						System.out.println("Error: Couldn't create directory '" + outPathChannel + "'");
						System.out.println("Skipping " + mediums[j]);
						continue;
					}
				}

				// Iterate through the logs.
				int users = 0;
				Enumeration userEnum = userLogs.keys();
				Vector userNames = new Vector();
	
				while (userEnum.hasMoreElements()) {
					users++;
					userNames.add((String)userEnum.nextElement());
				}
				Collections.sort(userNames);
	
				for (int u=0; u<users; u++) {
					int sessions = 0;
					int dupes = 0;
					int numFiles;
	
					String user = (String)userNames.elementAt(u);
					Vector logs = (Vector)userLogs.get(user);
	
					// Initialize; grab one session from each log, and record in log buffer
					for (int i=0; i<logs.size(); i++) {
						Log lg = (Log)logs.get(i);
						lg.sessDate = new Date(i); // initialize date, in case the readSession() func doesn't find a session date.
						if (!lg.readSession()) logs.remove(i--);
					}
					Collections.sort(logs);
					numFiles = logs.size();
					
					boolean useChDir = (j==IRC && user.indexOf("#") != -1) || (j==MSN && user.startsWith("Group Conversation"));
	
					File f = new File( ( useChDir ? outPath3Ch : outPath3 ) + File.separator + user + ".log");
					if (!f.exists()) f.createNewFile();
	
					try {
						OutputStreamWriter fw = new OutputStreamWriter(new FileOutputStream(f), "utf8");
						fw.write(UNICODEBOM);
						
						while (logs.size()>0) {
							sessions++;
							Log lg = (Log)logs.get(0);
							fw.write(lg.getSession());
	
							for (int i=1; i<logs.size(); i++) {
								Log olog = (Log)logs.get(i);
								if (!lg.sessDate.equals(olog.sessDate)) {
									//System.out.println("No Dup - " + lg.sessDate + " vs " + olog.sessDate);
									break;
								}
	
								dupes++;
								//System.out.println("-- Duplicate in " + lg.filename + "(" + lg.pathname + " & " + olog.pathname + ") on " + lg.sessDate);
								if (!olog.readSession()) logs.remove(i--);
							}
	
							if (!lg.readSession()) logs.remove(0);
							else Collections.sort(logs);
						}
	
						fw.close();
	
						System.out.println(user + ": merged " + numFiles + " files, " + sessions + " sessions, " + dupes + " duplicates");
					}
					catch (IOException e) {
						e.printStackTrace();
					}
	
				} //end for users
	
				System.out.println(users + " Log files written.\n");
			}


			/** ---- merge XML ---- **/

			// if XML files exist, make output dir
			if (userXMLs.size() > 0) {
				System.out.println("Starting merge, output format: Trillian XML\n");
				
				File outDir3 = new File(outPathQuery);
				if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
					System.out.println("Error: Couldn't create directory '" + outPathQuery + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}
				outDir3 = new File(outPathChannel);
				if (!outDir3.isDirectory() && !outDir3.mkdirs()) {
					System.out.println("Error: Couldn't create directory '" + outPathChannel + "'");
					System.out.println("Skipping " + mediums[j]);
					continue;
				}

				//Group Conversation

				// Iterate through the XML files.
				int users = 0;
				Enumeration userEnum = userXMLs.keys();
				Vector userNames = new Vector();
	
				while (userEnum.hasMoreElements()) {
					users++;
					userNames.add((String)userEnum.nextElement());
				}
				Collections.sort(userNames);
	
				for (int u=0; u<users; u++) {
					int sessions = 0;
					int dupes = 0;
					int numFiles;
					
					int numMinus = 0;
					int numMinusDupes = 0;
					Date minusDate = new Date(-1000);
	
					String user = (String)userNames.elementAt(u);
					Vector XMLs = (Vector)userXMLs.get(user);
	
					// Initialize; grab one session from each log, and record in log buffer
					for (int i=0; i<XMLs.size(); i++) {
						Log xml = (Log)XMLs.get(i);
						xml.sessDate = new Date( i - XMLs.size() - 1000); // initialize date, in case the readSession() func doesn't find a session date.
						if (!xml.readSession()) XMLs.remove(i--);
					}
					Collections.sort(XMLs);
					numFiles = XMLs.size();
	
					boolean useChDir = (j==IRC && user.indexOf("#") != -1) || (j==MSN && user.startsWith("Group Conversation"));

					File f = new File( ( useChDir ? outPathChannel : outPathQuery ) + File.separator + user + ".xml");
					if (!f.exists()) f.createNewFile();
	
					try {
						FileWriter fw = new FileWriter(f);
						//fw.write(UTF8BOM); // Trillian doesn't seem to UTF-8 the XML files
	
	
						while (XMLs.size()>0) {
							sessions++;
							Log xml = (Log)XMLs.get(0);
							fw.write(xml.getSession());
	
							if (xml.sessDate.equals(minusDate))
								numMinus++;
	
							for (int i=1; i<XMLs.size(); i++) {
								Log oxml = (Log)XMLs.get(i);
								if (!xml.sessDate.equals(oxml.sessDate) ) {
									//System.out.println("No Dup - " + xml.sessDate + " vs " + oxml.sessDate);
									break;
								}
	
								if (xml.sessDate.equals(minusDate)) {
									// if we find two sessions with -1, we can't be sure that they are equal, but we will treat them that way
									numMinusDupes++;
								}
								dupes++;
								//System.out.println("-- Duplicate in " + xml.filename + "(" + xml.pathname + " & " + oxml.pathname + ") on " + xml.sessDate);
								if (!oxml.readSession()) XMLs.remove(i--);
							}
	
							if (!xml.readSession()) XMLs.remove(0);
							else Collections.sort(XMLs);
						}
	
						fw.close();
	
						System.out.println(user + ": merged " + numFiles + " XML files, " + sessions + " sessions, " + dupes + " duplicates");
						if (numMinus > 0)
							System.out.println(" * Warning: kept " + numMinus + " sessions with a -1 date, an additional " + numMinusDupes + " may have been incorrectly dropped as duplicates");
					}
					catch (IOException e) {
						e.printStackTrace();
					}

				} //end while enumeration of users
	
				System.out.println(users + " XML files written.\n");
			}

			else {
				System.out.println("No files, skipping...\n");
				continue;
			}



		} //end for mediums

		in.close();

		System.out.println("Saving settings to " + settingsINI.getFileName());
		settingsINI.save();

		System.out.println("Finished.");

	} // end main

}
